# -*- coding: utf-8 -*-
#  Copyright (c) 2022, Markus Binsteiner
#
#  Mozilla Public License, version 2.0 (see LICENSE or https://www.mozilla.org/en-US/MPL/2.0/)

from kiara.doc.mkdocstrings.handler import get_handler

__all__ = ["get_handler"]  # noqa: WPS410
