# -*- coding: utf-8 -*-
# import rich_click as click

# from kiara import Kiara
# from kiara.service.rest import KiaraAPI


# @click.group()
# @click.pass_context
# def service(ctx):
#     """Service sub-commands.."""
#
#
# @service.command("start")
# @click.pass_context
# def start(ctx):
#
#     kiara: Kiara = ctx.obj["kiara"]
#
#     api = KiaraAPI(kiara=kiara)
#
#     api.start()
