# -*- coding: utf-8 -*-

#  Copyright (c) 2021, University of Luxembourg / DHARPA project
#
#  Mozilla Public License, version 2.0 (see LICENSE or https://www.mozilla.org/en-US/MPL/2.0/)


# def test_info():
#
#     runner = CliRunner()
#
#     result = runner.invoke(cli, "info --format json")
#
#     assert "table.query.sql" in result.stdout
#     json.loads(result.stdout)
